<?php

declare(strict_types=1);
/**
 * SPDX-FileCopyrightText: 2020 Nextcloud GmbH and Nextcloud contributors
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */
namespace Test;

use OC\EmojiHelper;
use OCP\IDBConnection;
use OCP\IEmojiHelper;

class EmojiHelperTest extends TestCase {
	/** @var IDBConnection|\PHPUnit\Framework\MockObject\MockObject */
	private $db;

	private IEmojiHelper $helper;

	protected function setUp(): void {
		parent::setUp();

		$this->db = $this->createMock(IDBConnection::class);
		$this->helper = new EmojiHelper($this->db);
	}

	/**
	 * @param bool $supports4ByteText
	 * @param bool $expected
	 *
	 * @dataProvider doesPlatformSupportEmojiDataProvider
	 */
	public function testDoesPlatformSupportEmoji(bool $supports4ByteText, bool $expected): void {
		$this->db->expects($this->once())
			->method('supports4ByteText')
			->willReturn($supports4ByteText);

		$this->assertEquals($expected, $this->helper->doesPlatformSupportEmoji());
	}

	/**
	 * @return array
	 */
	public function doesPlatformSupportEmojiDataProvider(): array {
		return [
			[true, true],
			[false, false],
		];
	}

	/**
	 * @param string $emoji
	 * @param bool $expected
	 *
	 * @dataProvider isValidSingleEmojiDataProvider
	 */
	public function testIsValidSingleEmoji(string $emoji, bool $expected): void {
		$actual = $this->helper->isValidSingleEmoji($emoji);

		$this->assertEquals($expected, $actual);
	}

	public function isValidSingleEmojiDataProvider(): array {
		return [
			['📱📠', false],
			['a', false],
			['0', false],
			['$', false],
			['$$', false],
			// Extracted from spreed/node_modules/emojis-list/index.js
			// https://github.com/Kikobeats/emojis-list/blob/master/index.js
			['🀄️', true],
			['🃏', true],
			['🅰️', true],
			['🅱️', true],
			['🅾️', true],
			['🅿️', true],
			['🆎', true],
			['🆑', true],
			['🆒', true],
			['🆓', true],
			['🆔', true],
			['🆕', true],
			['🆖', true],
			['🆗', true],
			['🆘', true],
			['🆙', true],
			['🆚', true],
			['🇦🇨', true],
			['🇦🇩', true],
			['🇦🇪', true],
			['🇦🇫', true],
			['🇦🇬', true],
			['🇦🇮', true],
			['🇦🇱', true],
			['🇦🇲', true],
			['🇦🇴', true],
			['🇦🇶', true],
			['🇦🇷', true],
			['🇦🇸', true],
			['🇦🇹', true],
			['🇦🇺', true],
			['🇦🇼', true],
			['🇦🇽', true],
			['🇦🇿', true],
			['🇦', true],
			['🇧🇦', true],
			['🇧🇧', true],
			['🇧🇩', true],
			['🇧🇪', true],
			['🇧🇫', true],
			['🇧🇬', true],
			['🇧🇭', true],
			['🇧🇮', true],
			['🇧🇯', true],
			['🇧🇱', true],
			['🇧🇲', true],
			['🇧🇳', true],
			['🇧🇴', true],
			['🇧🇶', true],
			['🇧🇷', true],
			['🇧🇸', true],
			['🇧🇹', true],
			['🇧🇻', true],
			['🇧🇼', true],
			['🇧🇾', true],
			['🇧🇿', true],
			['🇧', true],
			['🇨🇦', true],
			['🇨🇨', true],
			['🇨🇩', true],
			['🇨🇫', true],
			['🇨🇬', true],
			['🇨🇭', true],
			['🇨🇮', true],
			['🇨🇰', true],
			['🇨🇱', true],
			['🇨🇲', true],
			['🇨🇳', true],
			['🇨🇴', true],
			['🇨🇵', true],
			['🇨🇷', true],
			['🇨🇺', true],
			['🇨🇻', true],
			['🇨🇼', true],
			['🇨🇽', true],
			['🇨🇾', true],
			['🇨🇿', true],
			['🇨', true],
			['🇩🇪', true],
			['🇩🇬', true],
			['🇩🇯', true],
			['🇩🇰', true],
			['🇩🇲', true],
			['🇩🇴', true],
			['🇩🇿', true],
			['🇩', true],
			['🇪🇦', true],
			['🇪🇨', true],
			['🇪🇪', true],
			['🇪🇬', true],
			['🇪🇭', true],
			['🇪🇷', true],
			['🇪🇸', true],
			['🇪🇹', true],
			['🇪🇺', true],
			['🇪', true],
			['🇫🇮', true],
			['🇫🇯', true],
			['🇫🇰', true],
			['🇫🇲', true],
			['🇫🇴', true],
			['🇫🇷', true],
			['🇫', true],
			['🇬🇦', true],
			['🇬🇧', true],
			['🇬🇩', true],
			['🇬🇪', true],
			['🇬🇫', true],
			['🇬🇬', true],
			['🇬🇭', true],
			['🇬🇮', true],
			['🇬🇱', true],
			['🇬🇲', true],
			['🇬🇳', true],
			['🇬🇵', true],
			['🇬🇶', true],
			['🇬🇷', true],
			['🇬🇸', true],
			['🇬🇹', true],
			['🇬🇺', true],
			['🇬🇼', true],
			['🇬🇾', true],
			['🇬', true],
			['🇭🇰', true],
			['🇭🇲', true],
			['🇭🇳', true],
			['🇭🇷', true],
			['🇭🇹', true],
			['🇭🇺', true],
			['🇭', true],
			['🇮🇨', true],
			['🇮🇩', true],
			['🇮🇪', true],
			['🇮🇱', true],
			['🇮🇲', true],
			['🇮🇳', true],
			['🇮🇴', true],
			['🇮🇶', true],
			['🇮🇷', true],
			['🇮🇸', true],
			['🇮🇹', true],
			['🇮', true],
			['🇯🇪', true],
			['🇯🇲', true],
			['🇯🇴', true],
			['🇯🇵', true],
			['🇯', true],
			['🇰🇪', true],
			['🇰🇬', true],
			['🇰🇭', true],
			['🇰🇮', true],
			['🇰🇲', true],
			['🇰🇳', true],
			['🇰🇵', true],
			['🇰🇷', true],
			['🇰🇼', true],
			['🇰🇾', true],
			['🇰🇿', true],
			['🇰', true],
			['🇱🇦', true],
			['🇱🇧', true],
			['🇱🇨', true],
			['🇱🇮', true],
			['🇱🇰', true],
			['🇱🇷', true],
			['🇱🇸', true],
			['🇱🇹', true],
			['🇱🇺', true],
			['🇱🇻', true],
			['🇱🇾', true],
			['🇱', true],
			['🇲🇦', true],
			['🇲🇨', true],
			['🇲🇩', true],
			['🇲🇪', true],
			['🇲🇫', true],
			['🇲🇬', true],
			['🇲🇭', true],
			['🇲🇰', true],
			['🇲🇱', true],
			['🇲🇲', true],
			['🇲🇳', true],
			['🇲🇴', true],
			['🇲🇵', true],
			['🇲🇶', true],
			['🇲🇷', true],
			['🇲🇸', true],
			['🇲🇹', true],
			['🇲🇺', true],
			['🇲🇻', true],
			['🇲🇼', true],
			['🇲🇽', true],
			['🇲🇾', true],
			['🇲🇿', true],
			['🇲', true],
			['🇳🇦', true],
			['🇳🇨', true],
			['🇳🇪', true],
			['🇳🇫', true],
			['🇳🇬', true],
			['🇳🇮', true],
			['🇳🇱', true],
			['🇳🇴', true],
			['🇳🇵', true],
			['🇳🇷', true],
			['🇳🇺', true],
			['🇳🇿', true],
			['🇳', true],
			['🇴🇲', true],
			['🇴', true],
			['🇵🇦', true],
			['🇵🇪', true],
			['🇵🇫', true],
			['🇵🇬', true],
			['🇵🇭', true],
			['🇵🇰', true],
			['🇵🇱', true],
			['🇵🇲', true],
			['🇵🇳', true],
			['🇵🇷', true],
			['🇵🇸', true],
			['🇵🇹', true],
			['🇵🇼', true],
			['🇵🇾', true],
			['🇵', true],
			['🇶🇦', true],
			['🇶', true],
			['🇷🇪', true],
			['🇷🇴', true],
			['🇷🇸', true],
			['🇷🇺', true],
			['🇷🇼', true],
			['🇷', true],
			['🇸🇦', true],
			['🇸🇧', true],
			['🇸🇨', true],
			['🇸🇩', true],
			['🇸🇪', true],
			['🇸🇬', true],
			['🇸🇭', true],
			['🇸🇮', true],
			['🇸🇯', true],
			['🇸🇰', true],
			['🇸🇱', true],
			['🇸🇲', true],
			['🇸🇳', true],
			['🇸🇴', true],
			['🇸🇷', true],
			['🇸🇸', true],
			['🇸🇹', true],
			['🇸🇻', true],
			['🇸🇽', true],
			['🇸🇾', true],
			['🇸🇿', true],
			['🇸', true],
			['🇹🇦', true],
			['🇹🇨', true],
			['🇹🇩', true],
			['🇹🇫', true],
			['🇹🇬', true],
			['🇹🇭', true],
			['🇹🇯', true],
			['🇹🇰', true],
			['🇹🇱', true],
			['🇹🇲', true],
			['🇹🇳', true],
			['🇹🇴', true],
			['🇹🇷', true],
			['🇹🇹', true],
			['🇹🇻', true],
			['🇹🇼', true],
			['🇹🇿', true],
			['🇹', true],
			['🇺🇦', true],
			['🇺🇬', true],
			['🇺🇲', true],
			['🇺🇳', true],
			['🇺🇸', true],
			['🇺🇾', true],
			['🇺🇿', true],
			['🇺', true],
			['🇻🇦', true],
			['🇻🇨', true],
			['🇻🇪', true],
			['🇻🇬', true],
			['🇻🇮', true],
			['🇻🇳', true],
			['🇻🇺', true],
			['🇻', true],
			['🇼🇫', true],
			['🇼🇸', true],
			['🇼', true],
			['🇽🇰', true],
			['🇽', true],
			['🇾🇪', true],
			['🇾🇹', true],
			['🇾', true],
			['🇿🇦', true],
			['🇿🇲', true],
			['🇿🇼', true],
			['🇿', true],
			['🈁', true],
			['🈂️', true],
			['🈚️', true],
			['🈯️', true],
			['🈲', true],
			['🈳', true],
			['🈴', true],
			['🈵', true],
			['🈶', true],
			['🈷️', true],
			['🈸', true],
			['🈹', true],
			['🈺', true],
			['🉐', true],
			['🉑', true],
			['🌀', true],
			['🌁', true],
			['🌂', true],
			['🌃', true],
			['🌄', true],
			['🌅', true],
			['🌆', true],
			['🌇', true],
			['🌈', true],
			['🌉', true],
			['🌊', true],
			['🌋', true],
			['🌌', true],
			['🌍', true],
			['🌎', true],
			['🌏', true],
			['🌐', true],
			['🌑', true],
			['🌒', true],
			['🌓', true],
			['🌔', true],
			['🌕', true],
			['🌖', true],
			['🌗', true],
			['🌘', true],
			['🌙', true],
			['🌚', true],
			['🌛', true],
			['🌜', true],
			['🌝', true],
			['🌞', true],
			['🌟', true],
			['🌠', true],
			['🌡️', true],
			['🌤️', true],
			['🌥️', true],
			['🌦️', true],
			['🌧️', true],
			['🌨️', true],
			['🌩️', true],
			['🌪️', true],
			['🌫️', true],
			['🌬️', true],
			['🌭', true],
			['🌮', true],
			['🌯', true],
			['🌰', true],
			['🌱', true],
			['🌲', true],
			['🌳', true],
			['🌴', true],
			['🌵', true],
			['🌶️', true],
			['🌷', true],
			['🌸', true],
			['🌹', true],
			['🌺', true],
			['🌻', true],
			['🌼', true],
			['🌽', true],
			['🌾', true],
			['🌿', true],
			['🍀', true],
			['🍁', true],
			['🍂', true],
			['🍃', true],
			['🍄', true],
			['🍅', true],
			['🍆', true],
			['🍇', true],
			['🍈', true],
			['🍉', true],
			['🍊', true],
			['🍋', true],
			['🍌', true],
			['🍍', true],
			['🍎', true],
			['🍏', true],
			['🍐', true],
			['🍑', true],
			['🍒', true],
			['🍓', true],
			['🍔', true],
			['🍕', true],
			['🍖', true],
			['🍗', true],
			['🍘', true],
			['🍙', true],
			['🍚', true],
			['🍛', true],
			['🍜', true],
			['🍝', true],
			['🍞', true],
			['🍟', true],
			['🍠', true],
			['🍡', true],
			['🍢', true],
			['🍣', true],
			['🍤', true],
			['🍥', true],
			['🍦', true],
			['🍧', true],
			['🍨', true],
			['🍩', true],
			['🍪', true],
			['🍫', true],
			['🍬', true],
			['🍭', true],
			['🍮', true],
			['🍯', true],
			['🍰', true],
			['🍱', true],
			['🍲', true],
			['🍳', true],
			['🍴', true],
			['🍵', true],
			['🍶', true],
			['🍷', true],
			['🍸', true],
			['🍹', true],
			['🍺', true],
			['🍻', true],
			['🍼', true],
			['🍽️', true],
			['🍾', true],
			['🍿', true],
			['🎀', true],
			['🎁', true],
			['🎂', true],
			['🎃', true],
			['🎄', true],
			['🎅🏻', true],
			['🎅🏼', true],
			['🎅🏽', true],
			['🎅🏾', true],
			['🎅🏿', true],
			['🎅', true],
			['🎆', true],
			['🎇', true],
			['🎈', true],
			['🎉', true],
			['🎊', true],
			['🎋', true],
			['🎌', true],
			['🎍', true],
			['🎎', true],
			['🎏', true],
			['🎐', true],
			['🎑', true],
			['🎒', true],
			['🎓', true],
			['🎖️', true],
			['🎗️', true],
			['🎙️', true],
			['🎚️', true],
			['🎛️', true],
			['🎞️', true],
			['🎟️', true],
			['🎠', true],
			['🎡', true],
			['🎢', true],
			['🎣', true],
			['🎤', true],
			['🎥', true],
			['🎦', true],
			['🎧', true],
			['🎨', true],
			['🎩', true],
			['🎪', true],
			['🎫', true],
			['🎬', true],
			['🎭', true],
			['🎮', true],
			['🎯', true],
			['🎰', true],
			['🎱', true],
			['🎲', true],
			['🎳', true],
			['🎴', true],
			['🎵', true],
			['🎶', true],
			['🎷', true],
			['🎸', true],
			['🎹', true],
			['🎺', true],
			['🎻', true],
			['🎼', true],
			['🎽', true],
			['🎾', true],
			['🎿', true],
			['🏀', true],
			['🏁', true],
			['🏂🏻', true],
			['🏂🏼', true],
			['🏂🏽', true],
			['🏂🏾', true],
			['🏂🏿', true],
			['🏂', true],
			['🏃🏻‍♀️', true],
			['🏃🏻‍♂️', true],
			['🏃🏻', true],
			['🏃🏼‍♀️', true],
			['🏃🏼‍♂️', true],
			['🏃🏼', true],
			['🏃🏽‍♀️', true],
			['🏃🏽‍♂️', true],
			['🏃🏽', true],
			['🏃🏾‍♀️', true],
			['🏃🏾‍♂️', true],
			['🏃🏾', true],
			['🏃🏿‍♀️', true],
			['🏃🏿‍♂️', true],
			['🏃🏿', true],
			['🏃‍♀️', true],
			['🏃‍♂️', true],
			['🏃', true],
			['🏄🏻‍♀️', true],
			['🏄🏻‍♂️', true],
			['🏄🏻', true],
			['🏄🏼‍♀️', true],
			['🏄🏼‍♂️', true],
			['🏄🏼', true],
			['🏄🏽‍♀️', true],
			['🏄🏽‍♂️', true],
			['🏄🏽', true],
			['🏄🏾‍♀️', true],
			['🏄🏾‍♂️', true],
			['🏄🏾', true],
			['🏄🏿‍♀️', true],
			['🏄🏿‍♂️', true],
			['🏄🏿', true],
			['🏄‍♀️', true],
			['🏄‍♂️', true],
			['🏄', true],
			['🏅', true],
			['🏆', true],
			['🏇🏻', true],
			['🏇🏼', true],
			['🏇🏽', true],
			['🏇🏾', true],
			['🏇🏿', true],
			['🏇', true],
			['🏈', true],
			['🏉', true],
			['🏊🏻‍♀️', true],
			['🏊🏻‍♂️', true],
			['🏊🏻', true],
			['🏊🏼‍♀️', true],
			['🏊🏼‍♂️', true],
			['🏊🏼', true],
			['🏊🏽‍♀️', true],
			['🏊🏽‍♂️', true],
			['🏊🏽', true],
			['🏊🏾‍♀️', true],
			['🏊🏾‍♂️', true],
			['🏊🏾', true],
			['🏊🏿‍♀️', true],
			['🏊🏿‍♂️', true],
			['🏊🏿', true],
			['🏊‍♀️', true],
			['🏊‍♂️', true],
			['🏊', true],
			['🏋🏻‍♀️', true],
			['🏋🏻‍♂️', true],
			['🏋🏻', true],
			['🏋🏼‍♀️', true],
			['🏋🏼‍♂️', true],
			['🏋🏼', true],
			['🏋🏽‍♀️', true],
			['🏋🏽‍♂️', true],
			['🏋🏽', true],
			['🏋🏾‍♀️', true],
			['🏋🏾‍♂️', true],
			['🏋🏾', true],
			['🏋🏿‍♀️', true],
			['🏋🏿‍♂️', true],
			['🏋🏿', true],
			['🏋️‍♀️', true],
			['🏋️‍♂️', true],
			['🏋️', true],
			['🏌🏻‍♀️', true],
			['🏌🏻‍♂️', true],
			['🏌🏻', true],
			['🏌🏼‍♀️', true],
			['🏌🏼‍♂️', true],
			['🏌🏼', true],
			['🏌🏽‍♀️', true],
			['🏌🏽‍♂️', true],
			['🏌🏽', true],
			['🏌🏾‍♀️', true],
			['🏌🏾‍♂️', true],
			['🏌🏾', true],
			['🏌🏿‍♀️', true],
			['🏌🏿‍♂️', true],
			['🏌🏿', true],
			['🏌️‍♀️', true],
			['🏌️‍♂️', true],
			['🏌️', true],
			['🏍️', true],
			['🏎️', true],
			['🏏', true],
			['🏐', true],
			['🏑', true],
			['🏒', true],
			['🏓', true],
			['🏔️', true],
			['🏕️', true],
			['🏖️', true],
			['🏗️', true],
			['🏘️', true],
			['🏙️', true],
			['🏚️', true],
			['🏛️', true],
			['🏜️', true],
			['🏝️', true],
			['🏞️', true],
			['🏟️', true],
			['🏠', true],
			['🏡', true],
			['🏢', true],
			['🏣', true],
			['🏤', true],
			['🏥', true],
			['🏦', true],
			['🏧', true],
			['🏨', true],
			['🏩', true],
			['🏪', true],
			['🏫', true],
			['🏬', true],
			['🏭', true],
			['🏮', true],
			['🏯', true],
			['🏰', true],
			['🏳️‍🌈', true],
			['🏳️', true],
			['🏴‍☠️', true],
			['🏴󠁧󠁢󠁥󠁮󠁧󠁿', true],
			['🏴󠁧󠁢󠁳󠁣󠁴󠁿', true],
			['🏴󠁧󠁢󠁷󠁬󠁳󠁿', true],
			['🏴', true],
			['🏵️', true],
			['🏷️', true],
			['🏸', true],
			['🏹', true],
			['🏺', true],
			['🏻', true],
			['🏼', true],
			['🏽', true],
			['🏾', true],
			['🏿', true],
			['🐀', true],
			['🐁', true],
			['🐂', true],
			['🐃', true],
			['🐄', true],
			['🐅', true],
			['🐆', true],
			['🐇', true],
			['🐈', true],
			['🐉', true],
			['🐊', true],
			['🐋', true],
			['🐌', true],
			['🐍', true],
			['🐎', true],
			['🐏', true],
			['🐐', true],
			['🐑', true],
			['🐒', true],
			['🐓', true],
			['🐔', true],
			['🐕‍🦺', true],
			['🐕', true],
			['🐖', true],
			['🐗', true],
			['🐘', true],
			['🐙', true],
			['🐚', true],
			['🐛', true],
			['🐜', true],
			['🐝', true],
			['🐞', true],
			['🐟', true],
			['🐠', true],
			['🐡', true],
			['🐢', true],
			['🐣', true],
			['🐤', true],
			['🐥', true],
			['🐦', true],
			['🐧', true],
			['🐨', true],
			['🐩', true],
			['🐪', true],
			['🐫', true],
			['🐬', true],
			['🐭', true],
			['🐮', true],
			['🐯', true],
			['🐰', true],
			['🐱', true],
			['🐲', true],
			['🐳', true],
			['🐴', true],
			['🐵', true],
			['🐶', true],
			['🐷', true],
			['🐸', true],
			['🐹', true],
			['🐺', true],
			['🐻', true],
			['🐼', true],
			['🐽', true],
			['🐾', true],
			['🐿️', true],
			['👀', true],
			['👁‍🗨', true],
			['👁️', true],
			['👂🏻', true],
			['👂🏼', true],
			['👂🏽', true],
			['👂🏾', true],
			['👂🏿', true],
			['👂', true],
			['👃🏻', true],
			['👃🏼', true],
			['👃🏽', true],
			['👃🏾', true],
			['👃🏿', true],
			['👃', true],
			['👄', true],
			['👅', true],
			['👆🏻', true],
			['👆🏼', true],
			['👆🏽', true],
			['👆🏾', true],
			['👆🏿', true],
			['👆', true],
			['👇🏻', true],
			['👇🏼', true],
			['👇🏽', true],
			['👇🏾', true],
			['👇🏿', true],
			['👇', true],
			['👈🏻', true],
			['👈🏼', true],
			['👈🏽', true],
			['👈🏾', true],
			['👈🏿', true],
			['👈', true],
			['👉🏻', true],
			['👉🏼', true],
			['👉🏽', true],
			['👉🏾', true],
			['👉🏿', true],
			['👉', true],
			['👊🏻', true],
			['👊🏼', true],
			['👊🏽', true],
			['👊🏾', true],
			['👊🏿', true],
			['👊', true],
			['👋🏻', true],
			['👋🏼', true],
			['👋🏽', true],
			['👋🏾', true],
			['👋🏿', true],
			['👋', true],
			['👌🏻', true],
			['👌🏼', true],
			['👌🏽', true],
			['👌🏾', true],
			['👌🏿', true],
			['👌', true],
			['👍🏻', true],
			['👍🏼', true],
			['👍🏽', true],
			['👍🏾', true],
			['👍🏿', true],
			['👍', true],
			['👎🏻', true],
			['👎🏼', true],
			['👎🏽', true],
			['👎🏾', true],
			['👎🏿', true],
			['👎', true],
			['👏🏻', true],
			['👏🏼', true],
			['👏🏽', true],
			['👏🏾', true],
			['👏🏿', true],
			['👏', true],
			['👐🏻', true],
			['👐🏼', true],
			['👐🏽', true],
			['👐🏾', true],
			['👐🏿', true],
			['👐', true],
			['👑', true],
			['👒', true],
			['👓', true],
			['👔', true],
			['👕', true],
			['👖', true],
			['👗', true],
			['👘', true],
			['👙', true],
			['👚', true],
			['👛', true],
			['👜', true],
			['👝', true],
			['👞', true],
			['👟', true],
			['👠', true],
			['👡', true],
			['👢', true],
			['👣', true],
			['👤', true],
			['👥', true],
			['👦🏻', true],
			['👦🏼', true],
			['👦🏽', true],
			['👦🏾', true],
			['👦🏿', true],
			['👦', true],
			['👧🏻', true],
			['👧🏼', true],
			['👧🏽', true],
			['👧🏾', true],
			['👧🏿', true],
			['👧', true],
			['👨🏻‍🌾', true],
			['👨🏻‍🍳', true],
			['👨🏻‍🎓', true],
			['👨🏻‍🎤', true],
			['👨🏻‍🎨', true],
			['👨🏻‍🏫', true],
			['👨🏻‍🏭', true],
			['👨🏻‍💻', true],
			['👨🏻‍💼', true],
			['👨🏻‍🔧', true],
			['👨🏻‍🔬', true],
			['👨🏻‍🚀', true],
			['👨🏻‍🚒', true],
			['👨🏻‍🦯', true],
			['👨🏻‍🦰', true],
			['👨🏻‍🦱', true],
			['👨🏻‍🦲', true],
			['👨🏻‍🦳', true],
			['👨🏻‍🦼', true],
			['👨🏻‍🦽', true],
			['👨🏻‍⚕️', true],
			['👨🏻‍⚖️', true],
			['👨🏻‍✈️', true],
			['👨🏻', true],
			['👨🏼‍🌾', true],
			['👨🏼‍🍳', true],
			['👨🏼‍🎓', true],
			['👨🏼‍🎤', true],
			['👨🏼‍🎨', true],
			['👨🏼‍🏫', true],
			['👨🏼‍🏭', true],
			['👨🏼‍💻', true],
			['👨🏼‍💼', true],
			['👨🏼‍🔧', true],
			['👨🏼‍🔬', true],
			['👨🏼‍🚀', true],
			['👨🏼‍🚒', true],
			['👨🏼‍🤝‍👨🏻', true],
			['👨🏼‍🦯', true],
			['👨🏼‍🦰', true],
			['👨🏼‍🦱', true],
			['👨🏼‍🦲', true],
			['👨🏼‍🦳', true],
			['👨🏼‍🦼', true],
			['👨🏼‍🦽', true],
			['👨🏼‍⚕️', true],
			['👨🏼‍⚖️', true],
			['👨🏼‍✈️', true],
			['👨🏼', true],
			['👨🏽‍🌾', true],
			['👨🏽‍🍳', true],
			['👨🏽‍🎓', true],
			['👨🏽‍🎤', true],
			['👨🏽‍🎨', true],
			['👨🏽‍🏫', true],
			['👨🏽‍🏭', true],
			['👨🏽‍💻', true],
			['👨🏽‍💼', true],
			['👨🏽‍🔧', true],
			['👨🏽‍🔬', true],
			['👨🏽‍🚀', true],
			['👨🏽‍🚒', true],
			['👨🏽‍🤝‍👨🏻', true],
			['👨🏽‍🤝‍👨🏼', true],
			['👨🏽‍🦯', true],
			['👨🏽‍🦰', true],
			['👨🏽‍🦱', true],
			['👨🏽‍🦲', true],
			['👨🏽‍🦳', true],
			['👨🏽‍🦼', true],
			['👨🏽‍🦽', true],
			['👨🏽‍⚕️', true],
			['👨🏽‍⚖️', true],
			['👨🏽‍✈️', true],
			['👨🏽', true],
			['👨🏾‍🌾', true],
			['👨🏾‍🍳', true],
			['👨🏾‍🎓', true],
			['👨🏾‍🎤', true],
			['👨🏾‍🎨', true],
			['👨🏾‍🏫', true],
			['👨🏾‍🏭', true],
			['👨🏾‍💻', true],
			['👨🏾‍💼', true],
			['👨🏾‍🔧', true],
			['👨🏾‍🔬', true],
			['👨🏾‍🚀', true],
			['👨🏾‍🚒', true],
			['👨🏾‍🤝‍👨🏻', true],
			['👨🏾‍🤝‍👨🏼', true],
			['👨🏾‍🤝‍👨🏽', true],
			['👨🏾‍🦯', true],
			['👨🏾‍🦰', true],
			['👨🏾‍🦱', true],
			['👨🏾‍🦲', true],
			['👨🏾‍🦳', true],
			['👨🏾‍🦼', true],
			['👨🏾‍🦽', true],
			['👨🏾‍⚕️', true],
			['👨🏾‍⚖️', true],
			['👨🏾‍✈️', true],
			['👨🏾', true],
			['👨🏿‍🌾', true],
			['👨🏿‍🍳', true],
			['👨🏿‍🎓', true],
			['👨🏿‍🎤', true],
			['👨🏿‍🎨', true],
			['👨🏿‍🏫', true],
			['👨🏿‍🏭', true],
			['👨🏿‍💻', true],
			['👨🏿‍💼', true],
			['👨🏿‍🔧', true],
			['👨🏿‍🔬', true],
			['👨🏿‍🚀', true],
			['👨🏿‍🚒', true],
			['👨🏿‍🤝‍👨🏻', true],
			['👨🏿‍🤝‍👨🏼', true],
			['👨🏿‍🤝‍👨🏽', true],
			['👨🏿‍🤝‍👨🏾', true],
			['👨🏿‍🦯', true],
			['👨🏿‍🦰', true],
			['👨🏿‍🦱', true],
			['👨🏿‍🦲', true],
			['👨🏿‍🦳', true],
			['👨🏿‍🦼', true],
			['👨🏿‍🦽', true],
			['👨🏿‍⚕️', true],
			['👨🏿‍⚖️', true],
			['👨🏿‍✈️', true],
			['👨🏿', true],
			['👨‍🌾', true],
			['👨‍🍳', true],
			['👨‍🎓', true],
			['👨‍🎤', true],
			['👨‍🎨', true],
			['👨‍🏫', true],
			['👨‍🏭', true],
			['👨‍👦‍👦', true],
			['👨‍👦', true],
			['👨‍👧‍👦', true],
			['👨‍👧‍👧', true],
			['👨‍👧', true],
			['👨‍👨‍👦‍👦', true],
			['👨‍👨‍👦', true],
			['👨‍👨‍👧‍👦', true],
			['👨‍👨‍👧‍👧', true],
			['👨‍👨‍👧', true],
			['👨‍👩‍👦‍👦', true],
			['👨‍👩‍👦', true],
			['👨‍👩‍👧‍👦', true],
			['👨‍👩‍👧‍👧', true],
			['👨‍👩‍👧', true],
			['👨‍💻', true],
			['👨‍💼', true],
			['👨‍🔧', true],
			['👨‍🔬', true],
			['👨‍🚀', true],
			['👨‍🚒', true],
			['👨‍🦯', true],
			['👨‍🦰', true],
			['👨‍🦱', true],
			['👨‍🦲', true],
			['👨‍🦳', true],
			['👨‍🦼', true],
			['👨‍🦽', true],
			['👨‍⚕️', true],
			['👨‍⚖️', true],
			['👨‍✈️', true],
			['👨‍❤️‍👨', true],
			['👨‍❤️‍💋‍👨', true],
			['👨', true],
			['👩🏻‍🌾', true],
			['👩🏻‍🍳', true],
			['👩🏻‍🎓', true],
			['👩🏻‍🎤', true],
			['👩🏻‍🎨', true],
			['👩🏻‍🏫', true],
			['👩🏻‍🏭', true],
			['👩🏻‍💻', true],
			['👩🏻‍💼', true],
			['👩🏻‍🔧', true],
			['👩🏻‍🔬', true],
			['👩🏻‍🚀', true],
			['👩🏻‍🚒', true],
			['👩🏻‍🤝‍👨🏼', true],
			['👩🏻‍🤝‍👨🏽', true],
			['👩🏻‍🤝‍👨🏾', true],
			['👩🏻‍🤝‍👨🏿', true],
			['👩🏻‍🦯', true],
			['👩🏻‍🦰', true],
			['👩🏻‍🦱', true],
			['👩🏻‍🦲', true],
			['👩🏻‍🦳', true],
			['👩🏻‍🦼', true],
			['👩🏻‍🦽', true],
			['👩🏻‍⚕️', true],
			['👩🏻‍⚖️', true],
			['👩🏻‍✈️', true],
			['👩🏻', true],
			['👩🏼‍🌾', true],
			['👩🏼‍🍳', true],
			['👩🏼‍🎓', true],
			['👩🏼‍🎤', true],
			['👩🏼‍🎨', true],
			['👩🏼‍🏫', true],
			['👩🏼‍🏭', true],
			['👩🏼‍💻', true],
			['👩🏼‍💼', true],
			['👩🏼‍🔧', true],
			['👩🏼‍🔬', true],
			['👩🏼‍🚀', true],
			['👩🏼‍🚒', true],
			['👩🏼‍🤝‍👨🏻', true],
			['👩🏼‍🤝‍👨🏽', true],
			['👩🏼‍🤝‍👨🏾', true],
			['👩🏼‍🤝‍👨🏿', true],
			['👩🏼‍🤝‍👩🏻', true],
			['👩🏼‍🦯', true],
			['👩🏼‍🦰', true],
			['👩🏼‍🦱', true],
			['👩🏼‍🦲', true],
			['👩🏼‍🦳', true],
			['👩🏼‍🦼', true],
			['👩🏼‍🦽', true],
			['👩🏼‍⚕️', true],
			['👩🏼‍⚖️', true],
			['👩🏼‍✈️', true],
			['👩🏼', true],
			['👩🏽‍🌾', true],
			['👩🏽‍🍳', true],
			['👩🏽‍🎓', true],
			['👩🏽‍🎤', true],
			['👩🏽‍🎨', true],
			['👩🏽‍🏫', true],
			['👩🏽‍🏭', true],
			['👩🏽‍💻', true],
			['👩🏽‍💼', true],
			['👩🏽‍🔧', true],
			['👩🏽‍🔬', true],
			['👩🏽‍🚀', true],
			['👩🏽‍🚒', true],
			['👩🏽‍🤝‍👨🏻', true],
			['👩🏽‍🤝‍👨🏼', true],
			['👩🏽‍🤝‍👨🏾', true],
			['👩🏽‍🤝‍👨🏿', true],
			['👩🏽‍🤝‍👩🏻', true],
			['👩🏽‍🤝‍👩🏼', true],
			['👩🏽‍🦯', true],
			['👩🏽‍🦰', true],
			['👩🏽‍🦱', true],
			['👩🏽‍🦲', true],
			['👩🏽‍🦳', true],
			['👩🏽‍🦼', true],
			['👩🏽‍🦽', true],
			['👩🏽‍⚕️', true],
			['👩🏽‍⚖️', true],
			['👩🏽‍✈️', true],
			['👩🏽', true],
			['👩🏾‍🌾', true],
			['👩🏾‍🍳', true],
			['👩🏾‍🎓', true],
			['👩🏾‍🎤', true],
			['👩🏾‍🎨', true],
			['👩🏾‍🏫', true],
			['👩🏾‍🏭', true],
			['👩🏾‍💻', true],
			['👩🏾‍💼', true],
			['👩🏾‍🔧', true],
			['👩🏾‍🔬', true],
			['👩🏾‍🚀', true],
			['👩🏾‍🚒', true],
			['👩🏾‍🤝‍👨🏻', true],
			['👩🏾‍🤝‍👨🏼', true],
			['👩🏾‍🤝‍👨🏽', true],
			['👩🏾‍🤝‍👨🏿', true],
			['👩🏾‍🤝‍👩🏻', true],
			['👩🏾‍🤝‍👩🏼', true],
			['👩🏾‍🤝‍👩🏽', true],
			['👩🏾‍🦯', true],
			['👩🏾‍🦰', true],
			['👩🏾‍🦱', true],
			['👩🏾‍🦲', true],
			['👩🏾‍🦳', true],
			['👩🏾‍🦼', true],
			['👩🏾‍🦽', true],
			['👩🏾‍⚕️', true],
			['👩🏾‍⚖️', true],
			['👩🏾‍✈️', true],
			['👩🏾', true],
			['👩🏿‍🌾', true],
			['👩🏿‍🍳', true],
			['👩🏿‍🎓', true],
			['👩🏿‍🎤', true],
			['👩🏿‍🎨', true],
			['👩🏿‍🏫', true],
			['👩🏿‍🏭', true],
			['👩🏿‍💻', true],
			['👩🏿‍💼', true],
			['👩🏿‍🔧', true],
			['👩🏿‍🔬', true],
			['👩🏿‍🚀', true],
			['👩🏿‍🚒', true],
			['👩🏿‍🤝‍👨🏻', true],
			['👩🏿‍🤝‍👨🏼', true],
			['👩🏿‍🤝‍👨🏽', true],
			['👩🏿‍🤝‍👨🏾', true],
			['👩🏿‍🤝‍👩🏻', true],
			['👩🏿‍🤝‍👩🏼', true],
			['👩🏿‍🤝‍👩🏽', true],
			['👩🏿‍🤝‍👩🏾', true],
			['👩🏿‍🦯', true],
			['👩🏿‍🦰', true],
			['👩🏿‍🦱', true],
			['👩🏿‍🦲', true],
			['👩🏿‍🦳', true],
			['👩🏿‍🦼', true],
			['👩🏿‍🦽', true],
			['👩🏿‍⚕️', true],
			['👩🏿‍⚖️', true],
			['👩🏿‍✈️', true],
			['👩🏿', true],
			['👩‍🌾', true],
			['👩‍🍳', true],
			['👩‍🎓', true],
			['👩‍🎤', true],
			['👩‍🎨', true],
			['👩‍🏫', true],
			['👩‍🏭', true],
			['👩‍👦‍👦', true],
			['👩‍👦', true],
			['👩‍👧‍👦', true],
			['👩‍👧‍👧', true],
			['👩‍👧', true],
			['👩‍👩‍👦‍👦', true],
			['👩‍👩‍👦', true],
			['👩‍👩‍👧‍👦', true],
			['👩‍👩‍👧‍👧', true],
			['👩‍👩‍👧', true],
			['👩‍💻', true],
			['👩‍💼', true],
			['👩‍🔧', true],
			['👩‍🔬', true],
			['👩‍🚀', true],
			['👩‍🚒', true],
			['👩‍🦯', true],
			['👩‍🦰', true],
			['👩‍🦱', true],
			['👩‍🦲', true],
			['👩‍🦳', true],
			['👩‍🦼', true],
			['👩‍🦽', true],
			['👩‍⚕️', true],
			['👩‍⚖️', true],
			['👩‍✈️', true],
			['👩‍❤️‍👨', true],
			['👩‍❤️‍👩', true],
			['👩‍❤️‍💋‍👨', true],
			['👩‍❤️‍💋‍👩', true],
			['👩', true],
			['👪', true],
			['👫🏻', true],
			['👫🏼', true],
			['👫🏽', true],
			['👫🏾', true],
			['👫🏿', true],
			['👫', true],
			['👬🏻', true],
			['👬🏼', true],
			['👬🏽', true],
			['👬🏾', true],
			['👬🏿', true],
			['👬', true],
			['👭🏻', true],
			['👭🏼', true],
			['👭🏽', true],
			['👭🏾', true],
			['👭🏿', true],
			['👭', true],
			['👮🏻‍♀️', true],
			['👮🏻‍♂️', true],
			['👮🏻', true],
			['👮🏼‍♀️', true],
			['👮🏼‍♂️', true],
			['👮🏼', true],
			['👮🏽‍♀️', true],
			['👮🏽‍♂️', true],
			['👮🏽', true],
			['👮🏾‍♀️', true],
			['👮🏾‍♂️', true],
			['👮🏾', true],
			['👮🏿‍♀️', true],
			['👮🏿‍♂️', true],
			['👮🏿', true],
			['👮‍♀️', true],
			['👮‍♂️', true],
			['👮', true],
			['👯‍♀️', true],
			['👯‍♂️', true],
			['👯', true],
			['👰🏻', true],
			['👰🏼', true],
			['👰🏽', true],
			['👰🏾', true],
			['👰🏿', true],
			['👰', true],
			['👱🏻‍♀️', true],
			['👱🏻‍♂️', true],
			['👱🏻', true],
			['👱🏼‍♀️', true],
			['👱🏼‍♂️', true],
			['👱🏼', true],
			['👱🏽‍♀️', true],
			['👱🏽‍♂️', true],
			['👱🏽', true],
			['👱🏾‍♀️', true],
			['👱🏾‍♂️', true],
			['👱🏾', true],
			['👱🏿‍♀️', true],
			['👱🏿‍♂️', true],
			['👱🏿', true],
			['👱‍♀️', true],
			['👱‍♂️', true],
			['👱', true],
			['👲🏻', true],
			['👲🏼', true],
			['👲🏽', true],
			['👲🏾', true],
			['👲🏿', true],
			['👲', true],
			['👳🏻‍♀️', true],
			['👳🏻‍♂️', true],
			['👳🏻', true],
			['👳🏼‍♀️', true],
			['👳🏼‍♂️', true],
			['👳🏼', true],
			['👳🏽‍♀️', true],
			['👳🏽‍♂️', true],
			['👳🏽', true],
			['👳🏾‍♀️', true],
			['👳🏾‍♂️', true],
			['👳🏾', true],
			['👳🏿‍♀️', true],
			['👳🏿‍♂️', true],
			['👳🏿', true],
			['👳‍♀️', true],
			['👳‍♂️', true],
			['👳', true],
			['👴🏻', true],
			['👴🏼', true],
			['👴🏽', true],
			['👴🏾', true],
			['👴🏿', true],
			['👴', true],
			['👵🏻', true],
			['👵🏼', true],
			['👵🏽', true],
			['👵🏾', true],
			['👵🏿', true],
			['👵', true],
			['👶🏻', true],
			['👶🏼', true],
			['👶🏽', true],
			['👶🏾', true],
			['👶🏿', true],
			['👶', true],
			['👷🏻‍♀️', true],
			['👷🏻‍♂️', true],
			['👷🏻', true],
			['👷🏼‍♀️', true],
			['👷🏼‍♂️', true],
			['👷🏼', true],
			['👷🏽‍♀️', true],
			['👷🏽‍♂️', true],
			['👷🏽', true],
			['👷🏾‍♀️', true],
			['👷🏾‍♂️', true],
			['👷🏾', true],
			['👷🏿‍♀️', true],
			['👷🏿‍♂️', true],
			['👷🏿', true],
			['👷‍♀️', true],
			['👷‍♂️', true],
			['👷', true],
			['👸🏻', true],
			['👸🏼', true],
			['👸🏽', true],
			['👸🏾', true],
			['👸🏿', true],
			['👸', true],
			['👹', true],
			['👺', true],
			['👻', true],
			['👼🏻', true],
			['👼🏼', true],
			['👼🏽', true],
			['👼🏾', true],
			['👼🏿', true],
			['👼', true],
			['👽', true],
			['👾', true],
			['👿', true],
			['💀', true],
			['💁🏻‍♀️', true],
			['💁🏻‍♂️', true],
			['💁🏻', true],
			['💁🏼‍♀️', true],
			['💁🏼‍♂️', true],
			['💁🏼', true],
			['💁🏽‍♀️', true],
			['💁🏽‍♂️', true],
			['💁🏽', true],
			['💁🏾‍♀️', true],
			['💁🏾‍♂️', true],
			['💁🏾', true],
			['💁🏿‍♀️', true],
			['💁🏿‍♂️', true],
			['💁🏿', true],
			['💁‍♀️', true],
			['💁‍♂️', true],
			['💁', true],
			['💂🏻‍♀️', true],
			['💂🏻‍♂️', true],
			['💂🏻', true],
			['💂🏼‍♀️', true],
			['💂🏼‍♂️', true],
			['💂🏼', true],
			['💂🏽‍♀️', true],
			['💂🏽‍♂️', true],
			['💂🏽', true],
			['💂🏾‍♀️', true],
			['💂🏾‍♂️', true],
			['💂🏾', true],
			['💂🏿‍♀️', true],
			['💂🏿‍♂️', true],
			['💂🏿', true],
			['💂‍♀️', true],
			['💂‍♂️', true],
			['💂', true],
			['💃🏻', true],
			['💃🏼', true],
			['💃🏽', true],
			['💃🏾', true],
			['💃🏿', true],
			['💃', true],
			['💄', true],
			['💅🏻', true],
			['💅🏼', true],
			['💅🏽', true],
			['💅🏾', true],
			['💅🏿', true],
			['💅', true],
			['💆🏻‍♀️', true],
			['💆🏻‍♂️', true],
			['💆🏻', true],
			['💆🏼‍♀️', true],
			['💆🏼‍♂️', true],
			['💆🏼', true],
			['💆🏽‍♀️', true],
			['💆🏽‍♂️', true],
			['💆🏽', true],
			['💆🏾‍♀️', true],
			['💆🏾‍♂️', true],
			['💆🏾', true],
			['💆🏿‍♀️', true],
			['💆🏿‍♂️', true],
			['💆🏿', true],
			['💆‍♀️', true],
			['💆‍♂️', true],
			['💆', true],
			['💇🏻‍♀️', true],
			['💇🏻‍♂️', true],
			['💇🏻', true],
			['💇🏼‍♀️', true],
			['💇🏼‍♂️', true],
			['💇🏼', true],
			['💇🏽‍♀️', true],
			['💇🏽‍♂️', true],
			['💇🏽', true],
			['💇🏾‍♀️', true],
			['💇🏾‍♂️', true],
			['💇🏾', true],
			['💇🏿‍♀️', true],
			['💇🏿‍♂️', true],
			['💇🏿', true],
			['💇‍♀️', true],
			['💇‍♂️', true],
			['💇', true],
			['💈', true],
			['💉', true],
			['💊', true],
			['💋', true],
			['💌', true],
			['💍', true],
			['💎', true],
			['💏', true],
			['💐', true],
			['💑', true],
			['💒', true],
			['💓', true],
			['💔', true],
			['💕', true],
			['💖', true],
			['💗', true],
			['💘', true],
			['💙', true],
			['💚', true],
			['💛', true],
			['💜', true],
			['💝', true],
			['💞', true],
			['💟', true],
			['💠', true],
			['💡', true],
			['💢', true],
			['💣', true],
			['💤', true],
			['💥', true],
			['💦', true],
			['💧', true],
			['💨', true],
			['💩', true],
			['💪🏻', true],
			['💪🏼', true],
			['💪🏽', true],
			['💪🏾', true],
			['💪🏿', true],
			['💪', true],
			['💫', true],
			['💬', true],
			['💭', true],
			['💮', true],
			['💯', true],
			['💰', true],
			['💱', true],
			['💲', true],
			['💳', true],
			['💴', true],
			['💵', true],
			['💶', true],
			['💷', true],
			['💸', true],
			['💹', true],
			['💺', true],
			['💻', true],
			['💼', true],
			['💽', true],
			['💾', true],
			['💿', true],
			['📀', true],
			['📁', true],
			['📂', true],
			['📃', true],
			['📄', true],
			['📅', true],
			['📆', true],
			['📇', true],
			['📈', true],
			['📉', true],
			['📊', true],
			['📋', true],
			['📌', true],
			['📍', true],
			['📎', true],
			['📏', true],
			['📐', true],
			['📑', true],
			['📒', true],
			['📓', true],
			['📔', true],
			['📕', true],
			['📖', true],
			['📗', true],
			['📘', true],
			['📙', true],
			['📚', true],
			['📛', true],
			['📜', true],
			['📝', true],
			['📞', true],
			['📟', true],
			['📠', true],
			['📡', true],
			['📢', true],
			['📣', true],
			['📤', true],
			['📥', true],
			['📦', true],
			['📧', true],
			['📨', true],
			['📩', true],
			['📪', true],
			['📫', true],
			['📬', true],
			['📭', true],
			['📮', true],
			['📯', true],
			['📰', true],
			['📱', true],
			['📲', true],
			['📳', true],
			['📴', true],
			['📵', true],
			['📶', true],
			['📷', true],
			['📸', true],
			['📹', true],
			['📺', true],
			['📻', true],
			['📼', true],
			['📽️', true],
			['📿', true],
			['🔀', true],
			['🔁', true],
			['🔂', true],
			['🔃', true],
			['🔄', true],
			['🔅', true],
			['🔆', true],
			['🔇', true],
			['🔈', true],
			['🔉', true],
			['🔊', true],
			['🔋', true],
			['🔌', true],
			['🔍', true],
			['🔎', true],
			['🔏', true],
			['🔐', true],
			['🔑', true],
			['🔒', true],
			['🔓', true],
			['🔔', true],
			['🔕', true],
			['🔖', true],
			['🔗', true],
			['🔘', true],
			['🔙', true],
			['🔚', true],
			['🔛', true],
			['🔜', true],
			['🔝', true],
			['🔞', true],
			['🔟', true],
			['🔠', true],
			['🔡', true],
			['🔢', true],
			['🔣', true],
			['🔤', true],
			['🔥', true],
			['🔦', true],
			['🔧', true],
			['🔨', true],
			['🔩', true],
			['🔪', true],
			['🔫', true],
			['🔬', true],
			['🔭', true],
			['🔮', true],
			['🔯', true],
			['🔰', true],
			['🔱', true],
			['🔲', true],
			['🔳', true],
			['🔴', true],
			['🔵', true],
			['🔶', true],
			['🔷', true],
			['🔸', true],
			['🔹', true],
			['🔺', true],
			['🔻', true],
			['🔼', true],
			['🔽', true],
			['🕉️', true],
			['🕊️', true],
			['🕋', true],
			['🕌', true],
			['🕍', true],
			['🕎', true],
			['🕐', true],
			['🕑', true],
			['🕒', true],
			['🕓', true],
			['🕔', true],
			['🕕', true],
			['🕖', true],
			['🕗', true],
			['🕘', true],
			['🕙', true],
			['🕚', true],
			['🕛', true],
			['🕜', true],
			['🕝', true],
			['🕞', true],
			['🕟', true],
			['🕠', true],
			['🕡', true],
			['🕢', true],
			['🕣', true],
			['🕤', true],
			['🕥', true],
			['🕦', true],
			['🕧', true],
			['🕯️', true],
			['🕰️', true],
			['🕳️', true],
			['🕴🏻‍♀️', true],
			['🕴🏻‍♂️', true],
			['🕴🏻', true],
			['🕴🏼‍♀️', true],
			['🕴🏼‍♂️', true],
			['🕴🏼', true],
			['🕴🏽‍♀️', true],
			['🕴🏽‍♂️', true],
			['🕴🏽', true],
			['🕴🏾‍♀️', true],
			['🕴🏾‍♂️', true],
			['🕴🏾', true],
			['🕴🏿‍♀️', true],
			['🕴🏿‍♂️', true],
			['🕴🏿', true],
			['🕴️‍♀️', true],
			['🕴️‍♂️', true],
			['🕴️', true],
			['🕵🏻‍♀️', true],
			['🕵🏻‍♂️', true],
			['🕵🏻', true],
			['🕵🏼‍♀️', true],
			['🕵🏼‍♂️', true],
			['🕵🏼', true],
			['🕵🏽‍♀️', true],
			['🕵🏽‍♂️', true],
			['🕵🏽', true],
			['🕵🏾‍♀️', true],
			['🕵🏾‍♂️', true],
			['🕵🏾', true],
			['🕵🏿‍♀️', true],
			['🕵🏿‍♂️', true],
			['🕵🏿', true],
			['🕵️‍♀️', true],
			['🕵️‍♂️', true],
			['🕵️', true],
			['🕶️', true],
			['🕷️', true],
			['🕸️', true],
			['🕹️', true],
			['🕺🏻', true],
			['🕺🏼', true],
			['🕺🏽', true],
			['🕺🏾', true],
			['🕺🏿', true],
			['🕺', true],
			['🖇️', true],
			['🖊️', true],
			['🖋️', true],
			['🖌️', true],
			['🖍️', true],
			['🖐🏻', true],
			['🖐🏼', true],
			['🖐🏽', true],
			['🖐🏾', true],
			['🖐🏿', true],
			['🖐️', true],
			['🖕🏻', true],
			['🖕🏼', true],
			['🖕🏽', true],
			['🖕🏾', true],
			['🖕🏿', true],
			['🖕', true],
			['🖖🏻', true],
			['🖖🏼', true],
			['🖖🏽', true],
			['🖖🏾', true],
			['🖖🏿', true],
			['🖖', true],
			['🖤', true],
			['🖥️', true],
			['🖨️', true],
			['🖱️', true],
			['🖲️', true],
			['🖼️', true],
			['🗂️', true],
			['🗃️', true],
			['🗄️', true],
			['🗑️', true],
			['🗒️', true],
			['🗓️', true],
			['🗜️', true],
			['🗝️', true],
			['🗞️', true],
			['🗡️', true],
			['🗣️', true],
			['🗨️', true],
			['🗯️', true],
			['🗳️', true],
			['🗺️', true],
			['🗻', true],
			['🗼', true],
			['🗽', true],
			['🗾', true],
			['🗿', true],
			['😀', true],
			['😁', true],
			['😂', true],
			['😃', true],
			['😄', true],
			['😅', true],
			['😆', true],
			['😇', true],
			['😈', true],
			['😉', true],
			['😊', true],
			['😋', true],
			['😌', true],
			['😍', true],
			['😎', true],
			['😏', true],
			['😐', true],
			['😑', true],
			['😒', true],
			['😓', true],
			['😔', true],
			['😕', true],
			['😖', true],
			['😗', true],
			['😘', true],
			['😙', true],
			['😚', true],
			['😛', true],
			['😜', true],
			['😝', true],
			['😞', true],
			['😟', true],
			['😠', true],
			['😡', true],
			['😢', true],
			['😣', true],
			['😤', true],
			['😥', true],
			['😦', true],
			['😧', true],
			['😨', true],
			['😩', true],
			['😪', true],
			['😫', true],
			['😬', true],
			['😭', true],
			['😮', true],
			['😯', true],
			['😰', true],
			['😱', true],
			['😲', true],
			['😳', true],
			['😴', true],
			['😵', true],
			['😶', true],
			['😷', true],
			['😸', true],
			['😹', true],
			['😺', true],
			['😻', true],
			['😼', true],
			['😽', true],
			['😾', true],
			['😿', true],
			['🙀', true],
			['🙁', true],
			['🙂', true],
			['🙃', true],
			['🙄', true],
			['🙅🏻‍♀️', true],
			['🙅🏻‍♂️', true],
			['🙅🏻', true],
			['🙅🏼‍♀️', true],
			['🙅🏼‍♂️', true],
			['🙅🏼', true],
			['🙅🏽‍♀️', true],
			['🙅🏽‍♂️', true],
			['🙅🏽', true],
			['🙅🏾‍♀️', true],
			['🙅🏾‍♂️', true],
			['🙅🏾', true],
			['🙅🏿‍♀️', true],
			['🙅🏿‍♂️', true],
			['🙅🏿', true],
			['🙅‍♀️', true],
			['🙅‍♂️', true],
			['🙅', true],
			['🙆🏻‍♀️', true],
			['🙆🏻‍♂️', true],
			['🙆🏻', true],
			['🙆🏼‍♀️', true],
			['🙆🏼‍♂️', true],
			['🙆🏼', true],
			['🙆🏽‍♀️', true],
			['🙆🏽‍♂️', true],
			['🙆🏽', true],
			['🙆🏾‍♀️', true],
			['🙆🏾‍♂️', true],
			['🙆🏾', true],
			['🙆🏿‍♀️', true],
			['🙆🏿‍♂️', true],
			['🙆🏿', true],
			['🙆‍♀️', true],
			['🙆‍♂️', true],
			['🙆', true],
			['🙇🏻‍♀️', true],
			['🙇🏻‍♂️', true],
			['🙇🏻', true],
			['🙇🏼‍♀️', true],
			['🙇🏼‍♂️', true],
			['🙇🏼', true],
			['🙇🏽‍♀️', true],
			['🙇🏽‍♂️', true],
			['🙇🏽', true],
			['🙇🏾‍♀️', true],
			['🙇🏾‍♂️', true],
			['🙇🏾', true],
			['🙇🏿‍♀️', true],
			['🙇🏿‍♂️', true],
			['🙇🏿', true],
			['🙇‍♀️', true],
			['🙇‍♂️', true],
			['🙇', true],
			['🙈', true],
			['🙉', true],
			['🙊', true],
			['🙋🏻‍♀️', true],
			['🙋🏻‍♂️', true],
			['🙋🏻', true],
			['🙋🏼‍♀️', true],
			['🙋🏼‍♂️', true],
			['🙋🏼', true],
			['🙋🏽‍♀️', true],
			['🙋🏽‍♂️', true],
			['🙋🏽', true],
			['🙋🏾‍♀️', true],
			['🙋🏾‍♂️', true],
			['🙋🏾', true],
			['🙋🏿‍♀️', true],
			['🙋🏿‍♂️', true],
			['🙋🏿', true],
			['🙋‍♀️', true],
			['🙋‍♂️', true],
			['🙋', true],
			['🙌🏻', true],
			['🙌🏼', true],
			['🙌🏽', true],
			['🙌🏾', true],
			['🙌🏿', true],
			['🙌', true],
			['🙍🏻‍♀️', true],
			['🙍🏻‍♂️', true],
			['🙍🏻', true],
			['🙍🏼‍♀️', true],
			['🙍🏼‍♂️', true],
			['🙍🏼', true],
			['🙍🏽‍♀️', true],
			['🙍🏽‍♂️', true],
			['🙍🏽', true],
			['🙍🏾‍♀️', true],
			['🙍🏾‍♂️', true],
			['🙍🏾', true],
			['🙍🏿‍♀️', true],
			['🙍🏿‍♂️', true],
			['🙍🏿', true],
			['🙍‍♀️', true],
			['🙍‍♂️', true],
			['🙍', true],
			['🙎🏻‍♀️', true],
			['🙎🏻‍♂️', true],
			['🙎🏻', true],
			['🙎🏼‍♀️', true],
			['🙎🏼‍♂️', true],
			['🙎🏼', true],
			['🙎🏽‍♀️', true],
			['🙎🏽‍♂️', true],
			['🙎🏽', true],
			['🙎🏾‍♀️', true],
			['🙎🏾‍♂️', true],
			['🙎🏾', true],
			['🙎🏿‍♀️', true],
			['🙎🏿‍♂️', true],
			['🙎🏿', true],
			['🙎‍♀️', true],
			['🙎‍♂️', true],
			['🙎', true],
			['🙏🏻', true],
			['🙏🏼', true],
			['🙏🏽', true],
			['🙏🏾', true],
			['🙏🏿', true],
			['🙏', true],
			['🚀', true],
			['🚁', true],
			['🚂', true],
			['🚃', true],
			['🚄', true],
			['🚅', true],
			['🚆', true],
			['🚇', true],
			['🚈', true],
			['🚉', true],
			['🚊', true],
			['🚋', true],
			['🚌', true],
			['🚍', true],
			['🚎', true],
			['🚏', true],
			['🚐', true],
			['🚑', true],
			['🚒', true],
			['🚓', true],
			['🚔', true],
			['🚕', true],
			['🚖', true],
			['🚗', true],
			['🚘', true],
			['🚙', true],
			['🚚', true],
			['🚛', true],
			['🚜', true],
			['🚝', true],
			['🚞', true],
			['🚟', true],
			['🚠', true],
			['🚡', true],
			['🚢', true],
			['🚣🏻‍♀️', true],
			['🚣🏻‍♂️', true],
			['🚣🏻', true],
			['🚣🏼‍♀️', true],
			['🚣🏼‍♂️', true],
			['🚣🏼', true],
			['🚣🏽‍♀️', true],
			['🚣🏽‍♂️', true],
			['🚣🏽', true],
			['🚣🏾‍♀️', true],
			['🚣🏾‍♂️', true],
			['🚣🏾', true],
			['🚣🏿‍♀️', true],
			['🚣🏿‍♂️', true],
			['🚣🏿', true],
			['🚣‍♀️', true],
			['🚣‍♂️', true],
			['🚣', true],
			['🚤', true],
			['🚥', true],
			['🚦', true],
			['🚧', true],
			['🚨', true],
			['🚩', true],
			['🚪', true],
			['🚫', true],
			['🚬', true],
			['🚭', true],
			['🚮', true],
			['🚯', true],
			['🚰', true],
			['🚱', true],
			['🚲', true],
			['🚳', true],
			['🚴🏻‍♀️', true],
			['🚴🏻‍♂️', true],
			['🚴🏻', true],
			['🚴🏼‍♀️', true],
			['🚴🏼‍♂️', true],
			['🚴🏼', true],
			['🚴🏽‍♀️', true],
			['🚴🏽‍♂️', true],
			['🚴🏽', true],
			['🚴🏾‍♀️', true],
			['🚴🏾‍♂️', true],
			['🚴🏾', true],
			['🚴🏿‍♀️', true],
			['🚴🏿‍♂️', true],
			['🚴🏿', true],
			['🚴‍♀️', true],
			['🚴‍♂️', true],
			['🚴', true],
			['🚵🏻‍♀️', true],
			['🚵🏻‍♂️', true],
			['🚵🏻', true],
			['🚵🏼‍♀️', true],
			['🚵🏼‍♂️', true],
			['🚵🏼', true],
			['🚵🏽‍♀️', true],
			['🚵🏽‍♂️', true],
			['🚵🏽', true],
			['🚵🏾‍♀️', true],
			['🚵🏾‍♂️', true],
			['🚵🏾', true],
			['🚵🏿‍♀️', true],
			['🚵🏿‍♂️', true],
			['🚵🏿', true],
			['🚵‍♀️', true],
			['🚵‍♂️', true],
			['🚵', true],
			['🚶🏻‍♀️', true],
			['🚶🏻‍♂️', true],
			['🚶🏻', true],
			['🚶🏼‍♀️', true],
			['🚶🏼‍♂️', true],
			['🚶🏼', true],
			['🚶🏽‍♀️', true],
			['🚶🏽‍♂️', true],
			['🚶🏽', true],
			['🚶🏾‍♀️', true],
			['🚶🏾‍♂️', true],
			['🚶🏾', true],
			['🚶🏿‍♀️', true],
			['🚶🏿‍♂️', true],
			['🚶🏿', true],
			['🚶‍♀️', true],
			['🚶‍♂️', true],
			['🚶', true],
			['🚷', true],
			['🚸', true],
			['🚹', true],
			['🚺', true],
			['🚻', true],
			['🚼', true],
			['🚽', true],
			['🚾', true],
			['🚿', true],
			['🛀🏻', true],
			['🛀🏼', true],
			['🛀🏽', true],
			['🛀🏾', true],
			['🛀🏿', true],
			['🛀', true],
			['🛁', true],
			['🛂', true],
			['🛃', true],
			['🛄', true],
			['🛅', true],
			['🛋️', true],
			['🛌🏻', true],
			['🛌🏼', true],
			['🛌🏽', true],
			['🛌🏾', true],
			['🛌🏿', true],
			['🛌', true],
			['🛍️', true],
			['🛎️', true],
			['🛏️', true],
			['🛐', true],
			['🛑', true],
			['🛒', true],
			['🛕', true],
			['🛠️', true],
			['🛡️', true],
			['🛢️', true],
			['🛣️', true],
			['🛤️', true],
			['🛥️', true],
			['🛩️', true],
			['🛫', true],
			['🛬', true],
			['🛰️', true],
			['🛳️', true],
			['🛴', true],
			['🛵', true],
			['🛶', true],
			['🛷', true],
			['🛸', true],
			['🛹', true],
			['🛺', true],
			['🟠', true],
			['🟡', true],
			['🟢', true],
			['🟣', true],
			['🟤', true],
			['🟥', true],
			['🟦', true],
			['🟧', true],
			['🟨', true],
			['🟩', true],
			['🟪', true],
			['🟫', true],
			['🤍', true],
			['🤎', true],
			['🤏🏻', true],
			['🤏🏼', true],
			['🤏🏽', true],
			['🤏🏾', true],
			['🤏🏿', true],
			['🤏', true],
			['🤐', true],
			['🤑', true],
			['🤒', true],
			['🤓', true],
			['🤔', true],
			['🤕', true],
			['🤖', true],
			['🤗', true],
			['🤘🏻', true],
			['🤘🏼', true],
			['🤘🏽', true],
			['🤘🏾', true],
			['🤘🏿', true],
			['🤘', true],
			['🤙🏻', true],
			['🤙🏼', true],
			['🤙🏽', true],
			['🤙🏾', true],
			['🤙🏿', true],
			['🤙', true],
			['🤚🏻', true],
			['🤚🏼', true],
			['🤚🏽', true],
			['🤚🏾', true],
			['🤚🏿', true],
			['🤚', true],
			['🤛🏻', true],
			['🤛🏼', true],
			['🤛🏽', true],
			['🤛🏾', true],
			['🤛🏿', true],
			['🤛', true],
			['🤜🏻', true],
			['🤜🏼', true],
			['🤜🏽', true],
			['🤜🏾', true],
			['🤜🏿', true],
			['🤜', true],
			['🤝', true],
			['🤞🏻', true],
			['🤞🏼', true],
			['🤞🏽', true],
			['🤞🏾', true],
			['🤞🏿', true],
			['🤞', true],
			['🤟🏻', true],
			['🤟🏼', true],
			['🤟🏽', true],
			['🤟🏾', true],
			['🤟🏿', true],
			['🤟', true],
			['🤠', true],
			['🤡', true],
			['🤢', true],
			['🤣', true],
			['🤤', true],
			['🤥', true],
			['🤦🏻‍♀️', true],
			['🤦🏻‍♂️', true],
			['🤦🏻', true],
			['🤦🏼‍♀️', true],
			['🤦🏼‍♂️', true],
			['🤦🏼', true],
			['🤦🏽‍♀️', true],
			['🤦🏽‍♂️', true],
			['🤦🏽', true],
			['🤦🏾‍♀️', true],
			['🤦🏾‍♂️', true],
			['🤦🏾', true],
			['🤦🏿‍♀️', true],
			['🤦🏿‍♂️', true],
			['🤦🏿', true],
			['🤦‍♀️', true],
			['🤦‍♂️', true],
			['🤦', true],
			['🤧', true],
			['🤨', true],
			['🤩', true],
			['🤪', true],
			['🤫', true],
			['🤬', true],
			['🤭', true],
			['🤮', true],
			['🤯', true],
			['🤰🏻', true],
			['🤰🏼', true],
			['🤰🏽', true],
			['🤰🏾', true],
			['🤰🏿', true],
			['🤰', true],
			['🤱🏻', true],
			['🤱🏼', true],
			['🤱🏽', true],
			['🤱🏾', true],
			['🤱🏿', true],
			['🤱', true],
			['🤲🏻', true],
			['🤲🏼', true],
			['🤲🏽', true],
			['🤲🏾', true],
			['🤲🏿', true],
			['🤲', true],
			['🤳🏻', true],
			['🤳🏼', true],
			['🤳🏽', true],
			['🤳🏾', true],
			['🤳🏿', true],
			['🤳', true],
			['🤴🏻', true],
			['🤴🏼', true],
			['🤴🏽', true],
			['🤴🏾', true],
			['🤴🏿', true],
			['🤴', true],
			['🤵🏻‍♀️', true],
			['🤵🏻‍♂️', true],
			['🤵🏻', true],
			['🤵🏼‍♀️', true],
			['🤵🏼‍♂️', true],
			['🤵🏼', true],
			['🤵🏽‍♀️', true],
			['🤵🏽‍♂️', true],
			['🤵🏽', true],
			['🤵🏾‍♀️', true],
			['🤵🏾‍♂️', true],
			['🤵🏾', true],
			['🤵🏿‍♀️', true],
			['🤵🏿‍♂️', true],
			['🤵🏿', true],
			['🤵‍♀️', true],
			['🤵‍♂️', true],
			['🤵', true],
			['🤶🏻', true],
			['🤶🏼', true],
			['🤶🏽', true],
			['🤶🏾', true],
			['🤶🏿', true],
			['🤶', true],
			['🤷🏻‍♀️', true],
			['🤷🏻‍♂️', true],
			['🤷🏻', true],
			['🤷🏼‍♀️', true],
			['🤷🏼‍♂️', true],
			['🤷🏼', true],
			['🤷🏽‍♀️', true],
			['🤷🏽‍♂️', true],
			['🤷🏽', true],
			['🤷🏾‍♀️', true],
			['🤷🏾‍♂️', true],
			['🤷🏾', true],
			['🤷🏿‍♀️', true],
			['🤷🏿‍♂️', true],
			['🤷🏿', true],
			['🤷‍♀️', true],
			['🤷‍♂️', true],
			['🤷', true],
			['🤸🏻‍♀️', true],
			['🤸🏻‍♂️', true],
			['🤸🏻', true],
			['🤸🏼‍♀️', true],
			['🤸🏼‍♂️', true],
			['🤸🏼', true],
			['🤸🏽‍♀️', true],
			['🤸🏽‍♂️', true],
			['🤸🏽', true],
			['🤸🏾‍♀️', true],
			['🤸🏾‍♂️', true],
			['🤸🏾', true],
			['🤸🏿‍♀️', true],
			['🤸🏿‍♂️', true],
			['🤸🏿', true],
			['🤸‍♀️', true],
			['🤸‍♂️', true],
			['🤸', true],
			['🤹🏻‍♀️', true],
			['🤹🏻‍♂️', true],
			['🤹🏻', true],
			['🤹🏼‍♀️', true],
			['🤹🏼‍♂️', true],
			['🤹🏼', true],
			['🤹🏽‍♀️', true],
			['🤹🏽‍♂️', true],
			['🤹🏽', true],
			['🤹🏾‍♀️', true],
			['🤹🏾‍♂️', true],
			['🤹🏾', true],
			['🤹🏿‍♀️', true],
			['🤹🏿‍♂️', true],
			['🤹🏿', true],
			['🤹‍♀️', true],
			['🤹‍♂️', true],
			['🤹', true],
			['🤺', true],
			['🤼‍♀️', true],
			['🤼‍♂️', true],
			['🤼', true],
			['🤽🏻‍♀️', true],
			['🤽🏻‍♂️', true],
			['🤽🏻', true],
			['🤽🏼‍♀️', true],
			['🤽🏼‍♂️', true],
			['🤽🏼', true],
			['🤽🏽‍♀️', true],
			['🤽🏽‍♂️', true],
			['🤽🏽', true],
			['🤽🏾‍♀️', true],
			['🤽🏾‍♂️', true],
			['🤽🏾', true],
			['🤽🏿‍♀️', true],
			['🤽🏿‍♂️', true],
			['🤽🏿', true],
			['🤽‍♀️', true],
			['🤽‍♂️', true],
			['🤽', true],
			['🤾🏻‍♀️', true],
			['🤾🏻‍♂️', true],
			['🤾🏻', true],
			['🤾🏼‍♀️', true],
			['🤾🏼‍♂️', true],
			['🤾🏼', true],
			['🤾🏽‍♀️', true],
			['🤾🏽‍♂️', true],
			['🤾🏽', true],
			['🤾🏾‍♀️', true],
			['🤾🏾‍♂️', true],
			['🤾🏾', true],
			['🤾🏿‍♀️', true],
			['🤾🏿‍♂️', true],
			['🤾🏿', true],
			['🤾‍♀️', true],
			['🤾‍♂️', true],
			['🤾', true],
			['🤿', true],
			['🥀', true],
			['🥁', true],
			['🥂', true],
			['🥃', true],
			['🥄', true],
			['🥅', true],
			['🥇', true],
			['🥈', true],
			['🥉', true],
			['🥊', true],
			['🥋', true],
			['🥌', true],
			['🥍', true],
			['🥎', true],
			['🥏', true],
			['🥐', true],
			['🥑', true],
			['🥒', true],
			['🥓', true],
			['🥔', true],
			['🥕', true],
			['🥖', true],
			['🥗', true],
			['🥘', true],
			['🥙', true],
			['🥚', true],
			['🥛', true],
			['🥜', true],
			['🥝', true],
			['🥞', true],
			['🥟', true],
			['🥠', true],
			['🥡', true],
			['🥢', true],
			['🥣', true],
			['🥤', true],
			['🥥', true],
			['🥦', true],
			['🥧', true],
			['🥨', true],
			['🥩', true],
			['🥪', true],
			['🥫', true],
			['🥬', true],
			['🥭', true],
			['🥮', true],
			['🥯', true],
			['🥰', true],
			['🥱', true],
			['🥳', true],
			['🥴', true],
			['🥵', true],
			['🥶', true],
			['🥺', true],
			['🥻', true],
			['🥼', true],
			['🥽', true],
			['🥾', true],
			['🥿', true],
			['🦀', true],
			['🦁', true],
			['🦂', true],
			['🦃', true],
			['🦄', true],
			['🦅', true],
			['🦆', true],
			['🦇', true],
			['🦈', true],
			['🦉', true],
			['🦊', true],
			['🦋', true],
			['🦌', true],
			['🦍', true],
			['🦎', true],
			['🦏', true],
			['🦐', true],
			['🦑', true],
			['🦒', true],
			['🦓', true],
			['🦔', true],
			['🦕', true],
			['🦖', true],
			['🦗', true],
			['🦘', true],
			['🦙', true],
			['🦚', true],
			['🦛', true],
			['🦜', true],
			['🦝', true],
			['🦞', true],
			['🦟', true],
			['🦠', true],
			['🦡', true],
			['🦢', true],
			['🦥', true],
			['🦦', true],
			['🦧', true],
			['🦨', true],
			['🦩', true],
			['🦪', true],
			['🦮', true],
			['🦯', true],
			['🦰', true],
			['🦱', true],
			['🦲', true],
			['🦳', true],
			['🦴', true],
			['🦵🏻', true],
			['🦵🏼', true],
			['🦵🏽', true],
			['🦵🏾', true],
			['🦵🏿', true],
			['🦵', true],
			['🦶🏻', true],
			['🦶🏼', true],
			['🦶🏽', true],
			['🦶🏾', true],
			['🦶🏿', true],
			['🦶', true],
			['🦷', true],
			['🦸🏻‍♀️', true],
			['🦸🏻‍♂️', true],
			['🦸🏻', true],
			['🦸🏼‍♀️', true],
			['🦸🏼‍♂️', true],
			['🦸🏼', true],
			['🦸🏽‍♀️', true],
			['🦸🏽‍♂️', true],
			['🦸🏽', true],
			['🦸🏾‍♀️', true],
			['🦸🏾‍♂️', true],
			['🦸🏾', true],
			['🦸🏿‍♀️', true],
			['🦸🏿‍♂️', true],
			['🦸🏿', true],
			['🦸‍♀️', true],
			['🦸‍♂️', true],
			['🦸', true],
			['🦹🏻‍♀️', true],
			['🦹🏻‍♂️', true],
			['🦹🏻', true],
			['🦹🏼‍♀️', true],
			['🦹🏼‍♂️', true],
			['🦹🏼', true],
			['🦹🏽‍♀️', true],
			['🦹🏽‍♂️', true],
			['🦹🏽', true],
			['🦹🏾‍♀️', true],
			['🦹🏾‍♂️', true],
			['🦹🏾', true],
			['🦹🏿‍♀️', true],
			['🦹🏿‍♂️', true],
			['🦹🏿', true],
			['🦹‍♀️', true],
			['🦹‍♂️', true],
			['🦹', true],
			['🦺', true],
			['🦻🏻', true],
			['🦻🏼', true],
			['🦻🏽', true],
			['🦻🏾', true],
			['🦻🏿', true],
			['🦻', true],
			['🦼', true],
			['🦽', true],
			['🦾', true],
			['🦿', true],
			['🧀', true],
			['🧁', true],
			['🧂', true],
			['🧃', true],
			['🧄', true],
			['🧅', true],
			['🧆', true],
			['🧇', true],
			['🧈', true],
			['🧉', true],
			['🧊', true],
			['🧍🏻‍♀️', true],
			['🧍🏻‍♂️', true],
			['🧍🏻', true],
			['🧍🏼‍♀️', true],
			['🧍🏼‍♂️', true],
			['🧍🏼', true],
			['🧍🏽‍♀️', true],
			['🧍🏽‍♂️', true],
			['🧍🏽', true],
			['🧍🏾‍♀️', true],
			['🧍🏾‍♂️', true],
			['🧍🏾', true],
			['🧍🏿‍♀️', true],
			['🧍🏿‍♂️', true],
			['🧍🏿', true],
			['🧍‍♀️', true],
			['🧍‍♂️', true],
			['🧍', true],
			['🧎🏻‍♀️', true],
			['🧎🏻‍♂️', true],
			['🧎🏻', true],
			['🧎🏼‍♀️', true],
			['🧎🏼‍♂️', true],
			['🧎🏼', true],
			['🧎🏽‍♀️', true],
			['🧎🏽‍♂️', true],
			['🧎🏽', true],
			['🧎🏾‍♀️', true],
			['🧎🏾‍♂️', true],
			['🧎🏾', true],
			['🧎🏿‍♀️', true],
			['🧎🏿‍♂️', true],
			['🧎🏿', true],
			['🧎‍♀️', true],
			['🧎‍♂️', true],
			['🧎', true],
			['🧏🏻‍♀️', true],
			['🧏🏻‍♂️', true],
			['🧏🏻', true],
			['🧏🏼‍♀️', true],
			['🧏🏼‍♂️', true],
			['🧏🏼', true],
			['🧏🏽‍♀️', true],
			['🧏🏽‍♂️', true],
			['🧏🏽', true],
			['🧏🏾‍♀️', true],
			['🧏🏾‍♂️', true],
			['🧏🏾', true],
			['🧏🏿‍♀️', true],
			['🧏🏿‍♂️', true],
			['🧏🏿', true],
			['🧏‍♀️', true],
			['🧏‍♂️', true],
			['🧏', true],
			['🧐', true],
			['🧑🏻‍🤝‍🧑🏻', true],
			['🧑🏻', true],
			['🧑🏼‍🤝‍🧑🏻', true],
			['🧑🏼‍🤝‍🧑🏼', true],
			['🧑🏼', true],
			['🧑🏽‍🤝‍🧑🏻', true],
			['🧑🏽‍🤝‍🧑🏼', true],
			['🧑🏽‍🤝‍🧑🏽', true],
			['🧑🏽', true],
			['🧑🏾‍🤝‍🧑🏻', true],
			['🧑🏾‍🤝‍🧑🏼', true],
			['🧑🏾‍🤝‍🧑🏽', true],
			['🧑🏾‍🤝‍🧑🏾', true],
			['🧑🏾', true],
			['🧑🏿‍🤝‍🧑🏻', true],
			['🧑🏿‍🤝‍🧑🏼', true],
			['🧑🏿‍🤝‍🧑🏽', true],
			['🧑🏿‍🤝‍🧑🏾', true],
			['🧑🏿‍🤝‍🧑🏿', true],
			['🧑🏿', true],
			['🧑‍🤝‍🧑', true],
			['🧑', true],
			['🧒🏻', true],
			['🧒🏼', true],
			['🧒🏽', true],
			['🧒🏾', true],
			['🧒🏿', true],
			['🧒', true],
			['🧓🏻', true],
			['🧓🏼', true],
			['🧓🏽', true],
			['🧓🏾', true],
			['🧓🏿', true],
			['🧓', true],
			['🧔🏻', true],
			['🧔🏼', true],
			['🧔🏽', true],
			['🧔🏾', true],
			['🧔🏿', true],
			['🧔', true],
			['🧕🏻', true],
			['🧕🏼', true],
			['🧕🏽', true],
			['🧕🏾', true],
			['🧕🏿', true],
			['🧕', true],
			['🧖🏻‍♀️', true],
			['🧖🏻‍♂️', true],
			['🧖🏻', true],
			['🧖🏼‍♀️', true],
			['🧖🏼‍♂️', true],
			['🧖🏼', true],
			['🧖🏽‍♀️', true],
			['🧖🏽‍♂️', true],
			['🧖🏽', true],
			['🧖🏾‍♀️', true],
			['🧖🏾‍♂️', true],
			['🧖🏾', true],
			['🧖🏿‍♀️', true],
			['🧖🏿‍♂️', true],
			['🧖🏿', true],
			['🧖‍♀️', true],
			['🧖‍♂️', true],
			['🧖', true],
			['🧗🏻‍♀️', true],
			['🧗🏻‍♂️', true],
			['🧗🏻', true],
			['🧗🏼‍♀️', true],
			['🧗🏼‍♂️', true],
			['🧗🏼', true],
			['🧗🏽‍♀️', true],
			['🧗🏽‍♂️', true],
			['🧗🏽', true],
			['🧗🏾‍♀️', true],
			['🧗🏾‍♂️', true],
			['🧗🏾', true],
			['🧗🏿‍♀️', true],
			['🧗🏿‍♂️', true],
			['🧗🏿', true],
			['🧗‍♀️', true],
			['🧗‍♂️', true],
			['🧗', true],
			['🧘🏻‍♀️', true],
			['🧘🏻‍♂️', true],
			['🧘🏻', true],
			['🧘🏼‍♀️', true],
			['🧘🏼‍♂️', true],
			['🧘🏼', true],
			['🧘🏽‍♀️', true],
			['🧘🏽‍♂️', true],
			['🧘🏽', true],
			['🧘🏾‍♀️', true],
			['🧘🏾‍♂️', true],
			['🧘🏾', true],
			['🧘🏿‍♀️', true],
			['🧘🏿‍♂️', true],
			['🧘🏿', true],
			['🧘‍♀️', true],
			['🧘‍♂️', true],
			['🧘', true],
			['🧙🏻‍♀️', true],
			['🧙🏻‍♂️', true],
			['🧙🏻', true],
			['🧙🏼‍♀️', true],
			['🧙🏼‍♂️', true],
			['🧙🏼', true],
			['🧙🏽‍♀️', true],
			['🧙🏽‍♂️', true],
			['🧙🏽', true],
			['🧙🏾‍♀️', true],
			['🧙🏾‍♂️', true],
			['🧙🏾', true],
			['🧙🏿‍♀️', true],
			['🧙🏿‍♂️', true],
			['🧙🏿', true],
			['🧙‍♀️', true],
			['🧙‍♂️', true],
			['🧙', true],
			['🧚🏻‍♀️', true],
			['🧚🏻‍♂️', true],
			['🧚🏻', true],
			['🧚🏼‍♀️', true],
			['🧚🏼‍♂️', true],
			['🧚🏼', true],
			['🧚🏽‍♀️', true],
			['🧚🏽‍♂️', true],
			['🧚🏽', true],
			['🧚🏾‍♀️', true],
			['🧚🏾‍♂️', true],
			['🧚🏾', true],
			['🧚🏿‍♀️', true],
			['🧚🏿‍♂️', true],
			['🧚🏿', true],
			['🧚‍♀️', true],
			['🧚‍♂️', true],
			['🧚', true],
			['🧛🏻‍♀️', true],
			['🧛🏻‍♂️', true],
			['🧛🏻', true],
			['🧛🏼‍♀️', true],
			['🧛🏼‍♂️', true],
			['🧛🏼', true],
			['🧛🏽‍♀️', true],
			['🧛🏽‍♂️', true],
			['🧛🏽', true],
			['🧛🏾‍♀️', true],
			['🧛🏾‍♂️', true],
			['🧛🏾', true],
			['🧛🏿‍♀️', true],
			['🧛🏿‍♂️', true],
			['🧛🏿', true],
			['🧛‍♀️', true],
			['🧛‍♂️', true],
			['🧛', true],
			['🧜🏻‍♀️', true],
			['🧜🏻‍♂️', true],
			['🧜🏻', true],
			['🧜🏼‍♀️', true],
			['🧜🏼‍♂️', true],
			['🧜🏼', true],
			['🧜🏽‍♀️', true],
			['🧜🏽‍♂️', true],
			['🧜🏽', true],
			['🧜🏾‍♀️', true],
			['🧜🏾‍♂️', true],
			['🧜🏾', true],
			['🧜🏿‍♀️', true],
			['🧜🏿‍♂️', true],
			['🧜🏿', true],
			['🧜‍♀️', true],
			['🧜‍♂️', true],
			['🧜', true],
			['🧝🏻‍♀️', true],
			['🧝🏻‍♂️', true],
			['🧝🏻', true],
			['🧝🏼‍♀️', true],
			['🧝🏼‍♂️', true],
			['🧝🏼', true],
			['🧝🏽‍♀️', true],
			['🧝🏽‍♂️', true],
			['🧝🏽', true],
			['🧝🏾‍♀️', true],
			['🧝🏾‍♂️', true],
			['🧝🏾', true],
			['🧝🏿‍♀️', true],
			['🧝🏿‍♂️', true],
			['🧝🏿', true],
			['🧝‍♀️', true],
			['🧝‍♂️', true],
			['🧝', true],
			['🧞‍♀️', true],
			['🧞‍♂️', true],
			['🧞', true],
			['🧟‍♀️', true],
			['🧟‍♂️', true],
			['🧟', true],
			['🧠', true],
			['🧡', true],
			['🧢', true],
			['🧣', true],
			['🧤', true],
			['🧥', true],
			['🧦', true],
			['🧧', true],
			['🧨', true],
			['🧩', true],
			['🧪', true],
			['🧫', true],
			['🧬', true],
			['🧭', true],
			['🧮', true],
			['🧯', true],
			['🧰', true],
			['🧱', true],
			['🧲', true],
			['🧳', true],
			['🧴', true],
			['🧵', true],
			['🧶', true],
			['🧷', true],
			['🧸', true],
			['🧹', true],
			['🧺', true],
			['🧻', true],
			['🧼', true],
			['🧽', true],
			['🧾', true],
			['🧿', true],
			['🩰', true],
			['🩱', true],
			['🩲', true],
			['🩳', true],
			['🩸', true],
			['🩹', true],
			['🩺', true],
			['🪀', true],
			['🪁', true],
			['🪂', true],
			['🪐', true],
			['🪑', true],
			['🪒', true],
			['🪓', true],
			['🪔', true],
			['🪕', true],
			['‼️', true],
			['⁉️', true],
			['™️', true],
			['ℹ️', true],
			['↔️', true],
			['↕️', true],
			['↖️', true],
			['↗️', true],
			['↘️', true],
			['↙️', true],
			['↩️', true],
			['↪️', true],
			['#⃣', true],
			['⌚️', true],
			['⌛️', true],
			['⌨️', true],
			['⏏️', true],
			['⏩', true],
			['⏪', true],
			['⏫', true],
			['⏬', true],
			['⏭️', true],
			['⏮️', true],
			['⏯️', true],
			['⏰', true],
			['⏱️', true],
			['⏲️', true],
			['⏳', true],
			['⏸️', true],
			['⏹️', true],
			['⏺️', true],
			['Ⓜ️', true],
			['▪️', true],
			['▫️', true],
			['▶️', true],
			['◀️', true],
			['◻️', true],
			['◼️', true],
			['◽️', true],
			['◾️', true],
			['☀️', true],
			['☁️', true],
			['☂️', true],
			['☃️', true],
			['☄️', true],
			['☎️', true],
			['☑️', true],
			['☔️', true],
			['☕️', true],
			['☘️', true],
			['☝🏻', true],
			['☝🏼', true],
			['☝🏽', true],
			['☝🏾', true],
			['☝🏿', true],
			['☝️', true],
			['☠️', true],
			['☢️', true],
			['☣️', true],
			['☦️', true],
			['☪️', true],
			['☮️', true],
			['☯️', true],
			['☸️', true],
			['☹️', true],
			['☺️', true],
			['♀️', true],
			['♂️', true],
			['♈️', true],
			['♉️', true],
			['♊️', true],
			['♋️', true],
			['♌️', true],
			['♍️', true],
			['♎️', true],
			['♏️', true],
			['♐️', true],
			['♑️', true],
			['♒️', true],
			['♓️', true],
			['♟️', true],
			['♠️', true],
			['♣️', true],
			['♥️', true],
			['♦️', true],
			['♨️', true],
			['♻️', true],
			['♾', true],
			['♿️', true],
			['⚒️', true],
			['⚓️', true],
			['⚔️', true],
			['⚕️', true],
			['⚖️', true],
			['⚗️', true],
			['⚙️', true],
			['⚛️', true],
			['⚜️', true],
			['⚠️', true],
			['⚡️', true],
			['⚪️', true],
			['⚫️', true],
			['⚰️', true],
			['⚱️', true],
			['⚽️', true],
			['⚾️', true],
			['⛄️', true],
			['⛅️', true],
			['⛈️', true],
			['⛎', true],
			['⛏️', true],
			['⛑️', true],
			['⛓️', true],
			['⛔️', true],
			['⛩️', true],
			['⛪️', true],
			['⛰️', true],
			['⛱️', true],
			['⛲️', true],
			['⛳️', true],
			['⛴️', true],
			['⛵️', true],
			['⛷🏻', true],
			['⛷🏼', true],
			['⛷🏽', true],
			['⛷🏾', true],
			['⛷🏿', true],
			['⛷️', true],
			['⛸️', true],
			['⛹🏻‍♀️', true],
			['⛹🏻‍♂️', true],
			['⛹🏻', true],
			['⛹🏼‍♀️', true],
			['⛹🏼‍♂️', true],
			['⛹🏼', true],
			['⛹🏽‍♀️', true],
			['⛹🏽‍♂️', true],
			['⛹🏽', true],
			['⛹🏾‍♀️', true],
			['⛹🏾‍♂️', true],
			['⛹🏾', true],
			['⛹🏿‍♀️', true],
			['⛹🏿‍♂️', true],
			['⛹🏿', true],
			['⛹️‍♀️', true],
			['⛹️‍♂️', true],
			['⛹️', true],
			['⛺️', true],
			['⛽️', true],
			['✂️', true],
			['✅', true],
			['✈️', true],
			['✉️', true],
			['✊🏻', true],
			['✊🏼', true],
			['✊🏽', true],
			['✊🏾', true],
			['✊🏿', true],
			['✊', true],
			['✋🏻', true],
			['✋🏼', true],
			['✋🏽', true],
			['✋🏾', true],
			['✋🏿', true],
			['✋', true],
			['✌🏻', true],
			['✌🏼', true],
			['✌🏽', true],
			['✌🏾', true],
			['✌🏿', true],
			['✌️', true],
			['✍🏻', true],
			['✍🏼', true],
			['✍🏽', true],
			['✍🏾', true],
			['✍🏿', true],
			['✍️', true],
			['✏️', true],
			['✒️', true],
			['✔️', true],
			['✖️', true],
			['✝️', true],
			['✡️', true],
			['✨', true],
			['✳️', true],
			['✴️', true],
			['❄️', true],
			['❇️', true],
			['❌', true],
			['❎', true],
			['❓', true],
			['❔', true],
			['❕', true],
			['❗️', true],
			['❣️', true],
			['❤️', true],
			['➕', true],
			['➖', true],
			['➗', true],
			['➡️', true],
			['➰', true],
			['➿', true],
			['⤴️', true],
			['⤵️', true],
			['*⃣', true],
			['⬅️', true],
			['⬆️', true],
			['⬇️', true],
			['⬛️', true],
			['⬜️', true],
			['⭐️', true],
			['⭕️', true],
			['0⃣', true],
			['〰️', true],
			['〽️', true],
			['1⃣', true],
			['2⃣', true],
			['㊗️', true],
			['㊙️', true],
			['3⃣', true],
			['4⃣', true],
			['5⃣', true],
			['6⃣', true],
			['7⃣', true],
			['8⃣', true],
			['9⃣', true],
			['©️', true],
			['®️', true],
		];
	}
}
